# Copyright 2004 Tom Rothamel <tom-renpy@onegeek.org>
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
This module contains the code responsible for display management.
"""

import sys
import copy

import pygame
from pygame.locals import *

from params import config

# An event that is used to handle timer ticks.
TIMEREVENT = USEREVENT + 1

# The scene.
screen = None

class Transform(object):

    def __init__(self, t):
        self.x = 0
        self.y = 0
        self.time = t

    def move(self, x, y):
        rv = copy.copy(self)
        rv.x += x
        rv.y += y

        return rv

    def blit(self, src, x, y):
        screen.blit(src, (x + self.x, y + self.y))


class Showable(object):

    def key_event(self, e):
        return

    def mouse_event(self, e, x, y):
        return
    

def init():
    global screen
    pygame.init()
    screen = pygame.display.set_mode((config.screen_width, config.screen_height))

scene_list = None
time_start = 0

def reset_clock():
    time_start = pygame.time.get_ticks()

def clear():
    global scene_list
    scene_list = [ ]
    redraw(0)           

def show(s):
    scene_list.append(s)
    redraw(0)

def hide(s):
    scene_list.remove(s)
    redraw(0)

def draw():

    trans = Transform(pygame.time.get_ticks() - time_start)

    for s in scene_list:
        s.draw(trans)

    pygame.display.flip()

redraw_time = sys.maxint

def redraw(t):
    global redraw_time

    if t < redraw_time:
        redraw_time = t
    
def interact(ui):

    global redraw_time

    reverse_scene_list = scene_list[:]
    reverse_scene_list.reverse()
    
    keep_running = True

    pygame.event.set_allowed(TIMEREVENT)

    draw()

    while keep_running:

        if not pygame.event.peek([KEYDOWN,MOUSEMOTION,MOUSEBUTTONDOWN]):
            
            scenetime = pygame.time.get_ticks() - time_start
            timeleft = redraw_time - scenetime

            if timeleft <= 0:
                redraw_time = sys.maxint 
                draw()

            pygame.time.set_timer(TIMEREVENT, redraw_time - scenetime)

        e = pygame.event.wait()

        if e.type == TIMEREVENT:
            redraw_time = 0

        if e.type == QUIT:
            # Bail out. TODO make this better somehow.            
            pygame.quit()
            sys.exit(0)

        if e.type == KEYDOWN:

            # PULL
            if e.key == K_ESCAPE:
                ui.ui_escape()
                continue

            for s in reverse_scene_list:
                kerv = s.key_event(e)

                if kerv is True:
                    keep_running = False
                    break
                elif kerv is False:
                    break

        mouse = None

        if e.type == MOUSEMOTION:
            mouse = (False, e.pos[0], e.pos[1])

        if e.type == MOUSEBUTTONDOWN and e.button == 1:
            mouse = (True, e.pos[0], e.pos[1])

        if mouse:
            for s in reverse_scene_list:
                merv = s.mouse_event(*mouse)

                if merv is True:
                    keep_running = False
                    break
                elif merv is False:
                    break
            
            
