# Copyright 2004 Tom Rothamel <tom-renpy@onegeek.org>
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# Fragments of the game script.
import re

def nextify_list(l, next):
    """
    This iterates through each pair of elements l[n] and l[n+1] in
    a list of fragments, calling l[n].set_next(l[n+1]). Finally,
    this calls l[-1].set_next(next).
    
    It should be called by the set_next method of any fragment
    class that transfers control to a list.
    """
    
    for a, b in zip(l, l[1:]):
        a.set_next(b)

    l[-1].set_next(next)
    

# A map from fragment name to the fragment instance.
frag_dict = {}

# A list of all fragments known to the system.
frag_list = []

def finish(db, b):
    """
    Finishes loading fragments, by nextifying the fragment tree and
    placing every fragment into the fragment dictionary.
    """
    
    nextify_list(db, None)
    nextify_list(b, None)

    for f in frag_list:
        frag_dict[f.name] = f

def lookup(name):
    """
    Returns the fragment corresponding to name.
    """

    return frag_dict.get(name, None)

def fix_string(s):
    if s is None:
        return s

    # Strip off quote characters.
    s = s[1:-1]

    # Get rid of whitespace.
    s = re.sub(r"\s+", " ", s)

    replacements = {
        'n' : '\n',
        '\\' : '\\',
        't' : '    ',
        }

    
    def resub(m):
        c = m.group(1)
        if c in replacements:
            return replacements[c]
        else:
            return c

    s = s.strip()
    s = re.sub(r"\\(.)", resub, s)

    return s
    
class Fragment(object):
    """
    A fragment represents a portion of the script.

    @ivar next: The fragment that will be executed after this one is
    complete.

    @ivar name: The name of this fragment. This may be changed by the
    label command.
    """

    frag_serial = 0

    def __init__(self):
        self.name = "frag_%04d" % Fragment.frag_serial
        Fragment.frag_serial += 1
        
        self.next = None
        frag_list.append(self)

    def set_next(self, next):
        self.next = next

    def __repr__(self):
        return "<%s '%s'>" % (self.__class__.__name__, self.name)

    def run(self, e):
        print "Fragments of type %s not implemented." % self.__class__.__name__
        return self.next
        

# Control fragments.

class PythonFragment(Fragment):
    """
    This is a fragment that executes a block of python code.
    """
    
    def __init__(self, code):
        super(PythonFragment, self).__init__()        

        # We need to fix up the indentation of the python code.
        lines = [ ]
        indent = ""

        if code.startswith("\020"):
            code = code[1:-1]

        for l in code.split('\n'):
            while l.startswith("\020"):
                indent = indent + "    "
                l = l[1:]

            while l.startswith("\021"):
                indent = indent[:-4]
                l = l[1:]
            
            lines.append(indent + l)

        self.code = "\n".join(lines)

    def run(self, e):
        e.do_python(self.code)
        return self.next

class PassFragment(Fragment):

    def run(self, e):
        return self.next

class LabelFragment(Fragment):

    def __init__(self, name):
        super(LabelFragment, self).__init__()
        self.name = name

    def run(self, e):
        return self.next

class JumpFragment(Fragment):
    """
    A fragment that jumps to a target when executed, used to transfer
    control from one location to another in the program.
    """
    
    def __init__(self, next_name):
        super(JumpFragment, self).__init__()
        self.next_name = next_name

    def set_next(self, next):
        pass

    def run(self, e):
        return lookup(self.next_name)

class IfFragment(Fragment):
    """
    A fragment that evaluates a python expression, and then jumps to
    either a true or false branch, depending on the result.
    """

    def __init__(self, cond, true, false):
        super(IfFragment, self).__init__()

        self.cond = cond
        self.true = true
        self.false = false

    def set_next(self, next):
        nextify_list(self.true, next)

        if (self.false):
            nextify_list(self.false, next)
        else:
            self.next = next

    def run(self, e):
        if e.eval(self.cond):
            return self.true[0]

        if self.false:
            return self.false[0]
        else:
            return self.next
            
# UI Fragments

class DialogFragment(Fragment):
    """
    This contains a fragment of dialog in a script.

    @ivar who: The name of the character that is doing the speaking
    @ivar what: The dialog being spoken.
    """
    
    def __init__(self, who, what):
        super(DialogFragment, self).__init__()

        self.who = who # Will get fixed up by eval, if needed.
        self.what = fix_string(what)

    def run(self, e):
        e.do_dialog(self.who, self.what)
        return self.next

class MenuFragment(Fragment):
    """
    This fragment represents a menu that will be shown to the user.
    """

    def __init__(self, name, label):
        super(MenuFragment, self).__init__()

        if name:
            self.name = name

        self.label = fix_string(label)
        self.choices = []

    def add_choice(self, label, cmds):
        self.choices.append((fix_string(label), cmds))

    def set_next(self, next):
        for l, cmds in self.choices:
            nextify_list(cmds, next)

    def run(self, e):
        choice = e.do_menu(self.label, [ i[0] for i in self.choices ])
        return self.choices[choice][1][0]

# Declaration Fragments

class CharacterDeclarationFragment(Fragment):
    """
    This fragment represents the declaration of a character to the
    system.
    """

    def __init__(self, name):
        super(CharacterDeclarationFragment, self).__init__()
        self.character_name = name

    def run(self, e):
        e.do_character(self.character_name)
        return self.next

class ImageDeclarationFragment(Fragment):
    """
    This fragment represents the declaration of an image to the
    system.
    """

    def __init__(self, thing, img):
        super(ImageDeclarationFragment, self).__init__()
        
        self.thing = thing
        self.img = img

    def run(self, e):
        self.img.construct()
        e.ui_image(self.thing, self.img)
        return self.next

# Scene Fragments

class SceneFragment(Fragment):
    """
    This fragment represents the declaration of a new scene. It may
    also optionally contain a tuple describing something that will
    placed in that scene.
    """

    def __init__(self, thing):
        super(SceneFragment, self).__init__()

        self.thing = thing

    def run(self, e):
        e.do_scene()
        if self.thing:
            e.do_show(self.thing)

        return self.next
            

class ShowFragment(Fragment):
    """
    This fragment adds something to the scene.
    """

    def __init__(self, thing):
        super(ShowFragment, self).__init__()
        self.thing = thing

    def run(self, e):
        e.do_show(self.thing)

        return self.next
        

class HideFragment(Fragment):
    """
    This fragment removes something from the scene.
    """

    def __init__(self, thing):
        super(HideFragment, self).__init__()
        self.thing = thing
        
    def run(self, e):
        e.do_hide(self.thing)

        return self.next
