# Copyright 2004 Tom Rothamel <tom-renpy@onegeek.org>
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

# -*- tpg-mode -*-

language_grammar = r"""
set lexer = ContextSensitiveLexer

token ident : '\w+' ;
token NL : '\n' ;

token line : '[^\020\021\n]*' ;
token until_colon : '[^\020\021\n:]*' ;

token indent : '\020' ;
token dedent : '\021' ;

token string : '"(\\.|[^\\"])*"' ;

separator ws : '[ \t]+' ;
separator comment : '#[^\n]*' ;


START/rv -> 
        declaration_block/db
        block/b 
                $ finish(db, b) $
                $ rv = (db[0], b[0]) $
      ;

# Statements

block/l ->
        $ l = [] $
        ( NL
        | ( "label" | "def" ) ident/label ":" NL
                $ l.append(LabelFragment(label)) $
          ( indented_block/b  $ l.extend(b) $ )?
        | statement/s $ l.append(s) $
        )+
        ;

indented_block/l ->
        ( NL )*
        indent
        block/l
        dedent
        ;

indented_block_or_statement/l ->
        ( statement/s  $ l = [ s ] $
        | NL indented_block/l
        )
        ;

python_block/s ->
        indent $ s = "\020" $
        ( python_block/ss $ s += ss $
        | line/l NL $ s += l + "\n" $ 
        )*
        dedent $ s += "\21" $
        ;

statement_or_declaration/s ->
        (  # Python code.

           "\$" line/s NL 
                $ s = PythonFragment(s) $        
        | "python" ":" NL python_block/s 
                $ s = PythonFragment(s) $ 
        )
        ;
        

# A thing that can be in a scene.
thing/t ->
        ident/a
        $ t = [ a ] $
        ( ident/b  $ t.append(b) $ )*
        $ t = tuple(t) $
        ;
        
statement/s ->
        ( statement_or_declaration/s

          # Control flow.

        | "if" until_colon/cond ':' indented_block_or_statement/true
                $ false = None $
          ( "else" ":" indented_block_or_statement/false )?
                $ s = IfFragment(cond, true, false) $

        | "jump" ident/l NL
                $ s = JumpFragment(l) $

        | "pass" NL
                $ s = PassFragment() $

          # Scene management.

        | "scene" ( thing/t | $ t = None $ ) NL
                $ s = SceneFragment(t) $

        | "show" thing/t NL
                $ s = ShowFragment(t) $ 

        | "hide" thing/t NL
                $ s = HideFragment(t) $ 
        
          # Menus and dialog.

        | "menu" ( ident/name | $ name = None $ ) ":" NL
          indent
          ( string/label NL 
          | $ label = None $
          )
                $ s = MenuFragment(name, label) $
          ( string/label ':' indented_block_or_statement/choice 
                $ s.add_choice(label, choice) $
          | NL
          )+
          dedent        

        | ident/who string/what NL 
                $ s = DialogFragment(who, what) $
        | string/who string/what NL
                $ s = DialogFragment(who, what) $
        | string/what NL
                $ s = DialogFragment(None, what) $
        )
        ;

# Declarations

declaration_block/l ->
        $ l = [] $
        ( NL
        | declaration/d $ l.append(d) $
        )*
        ;

image/img ->
        ( string/s $ img = image.image(s[1:-1]) $
        )
        ;

declaration/d ->
        ( statement_or_declaration/d
        
        | "character" ident/name NL
                $ d = CharacterDeclarationFragment(name) $
        | "image" thing/t ":" image/img NL 
                $ d = ImageDeclarationFragment(t, img) $
        )
        ;
"""

indent_grammar = r"""
set lexer = ContextSensitiveLexer

# This must be synced up with language.pyg
token string : '"(\\.|[^\\"])*"';
token nl : '\n' ;
token ignored_indentation : ' *(?=[#\n])' ;
token indentation : ' *' ;
token rest : '[^\n"]+' ;

START -> 
        ( string/s $ self.accum(s) $
        | nl/s $ self.accum(s) $ 
            ( ignored_indentation/s $ self.accum(s) $
            | indentation/s $ self.indent_dedent(s) $ 
            )
        | rest/s $ self.accum(s) $ 
        )*
        ;
"""